/* GTK - The GIMP Toolkit
 * gtkcupssecretsutils.h: Helper to use a secrets service for printer passwords
 * Copyright (C) 2014, Intevation GmbH
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib.h>
#include <gio/gio.h>
#include <string.h>

#include <gtk/gtk.h>

#define SECRETS_BUS "org.freedesktop.secrets"
#define SECRETS_IFACE(interface) "org.freedesktop.Secret."interface
#define SECRETS_PATH "/org/freedesktop/secrets"
#define SECRETS_TIMEOUT 5000

static gboolean
open_session (GDBusConnection **session_connection,
              GDBusProxy **service_proxy,
              GDBusProxy **session_proxy)
{
  GError *error = NULL;
  GVariant *result = NULL,
           *subresult = NULL;

  /* Connect to the secret service interface */
  *session_connection = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
  if (!*session_connection)
    {
      GTK_NOTE (PRINTING,
                g_print ("Failed to open dbus session: %s.\n", error->message));
      return FALSE;
    }

  *service_proxy = g_dbus_proxy_new_sync (*session_connection,
                                          G_DBUS_PROXY_FLAGS_NONE,
                                          NULL,
                                          SECRETS_BUS,
                                          SECRETS_PATH,
                                          SECRETS_IFACE("Service"),
                                          NULL,
                                          &error);
  if (!*service_proxy)
    {
      GTK_NOTE (PRINTING,
                g_print ("Failed to obtain service interface: %s.\n", error->message));
      goto done;
    }

  /* Open a session */
  result = g_dbus_proxy_call_sync (*service_proxy,
                                   "OpenSession",
                                   g_variant_new("(sv)", "plain",
                                     g_variant_new_string("")),
                                   G_DBUS_CALL_FLAGS_NONE,
                                   SECRETS_TIMEOUT,
                                   NULL,
                                   &error);
  if (!result)
    {
      GTK_NOTE (PRINTING,
                g_print ("Failed to open secrets session: %s.\n", error->message));
      goto done;
    }

  subresult = g_variant_get_child_value (result, 1);
  g_variant_unref(result);
  result = NULL;

  if (!subresult)
    {
      GTK_NOTE (PRINTING, g_print ("Invalid session path response.\n"));
      goto done;
    }

  const gchar *const session_path = g_variant_get_string(subresult, NULL);

  *session_proxy = g_dbus_proxy_new_sync(*session_connection,
                                        G_DBUS_PROXY_FLAGS_NONE,
                                        NULL,
                                        SECRETS_BUS,
                                        session_path,
                                        SECRETS_IFACE("Session"),
                                        NULL,
                                        &error);
  if (!*session_proxy)
    {
      GTK_NOTE (PRINTING,
                g_print ("Could not obtain session proxy: %s.\n", error->message));
      goto done;
    }
  g_variant_unref (subresult);
  subresult = NULL;

done:
  if (*session_connection && *session_proxy && *service_proxy && !error)
      return TRUE;

  /* Cleanup in case of errors */
  if (error)
      g_error_free (error);

  if (*session_proxy)
    {
      g_dbus_proxy_call_sync (*session_proxy,
                              "Close",
                              g_variant_new("()"),
                              G_DBUS_CALL_FLAGS_NONE,
                              SECRETS_TIMEOUT,
                              NULL,
                              NULL);
      g_object_unref (*session_proxy);
    }

  if (service_proxy)
      g_object_unref (service_proxy);

  if (*session_connection)
      g_object_unref (*session_connection);

  return FALSE;
}

static void
close_session (GDBusConnection *session_connection,
               GDBusProxy *service_proxy,
               GDBusProxy *session_proxy)
{
  if (service_proxy)
      g_object_unref (service_proxy);

  if (session_proxy)
    {
      g_dbus_proxy_call_sync (session_proxy,
                              "Close",
                              g_variant_new("()"),
                              G_DBUS_CALL_FLAGS_NONE,
                              SECRETS_TIMEOUT,
                              NULL,
                              NULL);
      g_object_unref (session_proxy);
    }

  if (session_connection)
      g_object_unref (session_connection);
}


/**
 * secrets_service_available:
 *
 * Check if a secrets service is available.
 *
 * Returns: True if a secret service interface could be opened.
 */
gboolean
secrets_service_available ()
{
  GDBusConnection *session_connection = NULL;
  GDBusProxy *service_proxy = NULL;
  session_connection = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, NULL);

  if (!session_connection)
      return FALSE;

  service_proxy = g_dbus_proxy_new_sync (session_connection,
                                         G_DBUS_PROXY_FLAGS_NONE,
                                         NULL,
                                         SECRETS_BUS,
                                         SECRETS_PATH,
                                         SECRETS_IFACE("Service"),
                                         NULL,
                                         NULL);
  g_object_unref (session_connection);
  if (!service_proxy)
      return FALSE;
  return TRUE;
}


static gboolean
unlock_collection (GDBusConnection *session_connection,
                   GDBusProxy *service_proxy,
                   GDBusProxy *collection_proxy)
{
  GError *error = NULL;
  GVariant *result = NULL;
  const gchar *prompt_path;

  const gchar * const to_unlock[] = {
    g_dbus_proxy_get_object_path(collection_proxy),
    NULL
  };

  /* Try to unlock the collection */
  result = g_dbus_proxy_call_sync (service_proxy,
                                   "Unlock",
                                   g_variant_new("(^ao)", to_unlock),
                                   G_DBUS_CALL_FLAGS_NONE,
                                   SECRETS_TIMEOUT,
                                   NULL,
                                   &error);
  if (!result)
    {
      GTK_NOTE (PRINTING,
                g_print ("Failed to unlock: %s.\n", error->message));
      g_error_free (error);
      return FALSE;
    }

  g_variant_get (result, "(@ao&o)", NULL, &prompt_path);

  if (prompt_path && strlen (prompt_path) > 2)
    {
      /* The user must be prompted to unlock the collection */
      GDBusProxy *prompt_proxy = g_dbus_proxy_new_sync(session_connection,
                                                       G_DBUS_PROXY_FLAGS_NONE,
                                                       NULL,
                                                       SECRETS_BUS,
                                                       prompt_path,
                                                       SECRETS_IFACE("Prompt"),
                                                       NULL,
                                                       &error);
      if (!prompt_proxy)
        {
          GTK_NOTE (PRINTING,
                    g_print ("Failed to access prompt interface: %s.\n",
                             error->message));
          g_error_free (error);
          return FALSE;
        }
      result = g_dbus_proxy_call_sync (prompt_proxy,
                                       "Prompt",
                                       g_variant_new("(s)", "0"),
                                       G_DBUS_CALL_FLAGS_NONE,
                                       SECRETS_TIMEOUT,
                                       NULL,
                                       &error);
      if (!result)
        {
          GTK_NOTE (PRINTING,
                    g_print ("Failed to prompt for password: %s.\n",
                             error->message));
          g_error_free (error);
          return FALSE;
        }
      g_variant_unref (result);
      result = NULL;
      g_object_unref (prompt_proxy);
    }

    return TRUE;
}

static GDBusProxy *
get_default_collection_unlocked (GDBusConnection *session_connection,
                                 GDBusProxy *service_proxy)
{
  GError *error = NULL;
  GDBusProxy *collection_proxy = NULL;
  const gchar *collection_path;
  GVariant *locked = NULL,
           *result = NULL,
           *subresult = NULL;
  gboolean is_locked = FALSE;
  gsize path_len = 0;

  /* Look up default collection */
  result = g_dbus_proxy_call_sync (service_proxy,
                                   "ReadAlias",
                                   g_variant_new("(s)", "default"),
                                   G_DBUS_CALL_FLAGS_NONE,
                                   SECRETS_TIMEOUT,
                                   NULL,
                                   &error);
  if (!result)
    {
      GTK_NOTE (PRINTING,
                g_print ("Failed to obtain default collection name: %s.\n",
                         error->message));
      goto done;
    }

  subresult = g_variant_get_child_value (result, 0);
  g_variant_unref (result);
  result = NULL;

  if (!subresult)
    {
      GTK_NOTE (PRINTING, g_print ("Invalid ReadAlias response.\n"));
      goto done;
    }

  collection_path = g_variant_get_string(subresult, &path_len);

  if (!collection_path || path_len <= 1)
    {
      /* If there is no default collection collection_path is "/".
       * It might be reasonable to create the collection in this case.
       */
      goto done;
    }

  collection_proxy = g_dbus_proxy_new_sync (session_connection,
                                            G_DBUS_PROXY_FLAGS_NONE,
                                            NULL,
                                            SECRETS_BUS,
                                            collection_path,
                                            SECRETS_IFACE("Collection"),
                                            NULL,
                                            &error);
  if (!collection_proxy)
    {
      GTK_NOTE (PRINTING,
                g_print ("Failed to access collection interface: %s.\n",
                         error->message));
      goto done;
    }
  g_variant_unref (subresult);
  subresult = NULL;

  locked = g_dbus_proxy_get_cached_property (collection_proxy,
                                             "Locked");
  if (!locked)
    {
      GTK_NOTE (PRINTING,
                g_print ("Failed to lookup collection locked property.\n"));
      goto done;
    }
  g_variant_get (locked, "b", &is_locked);
  g_variant_unref (locked);

  if (is_locked)
     is_locked = unlock_collection(session_connection, service_proxy,
                                   collection_proxy);

done:
  /* Cleanup and exit */
  if (result)
      g_variant_unref (result);
  if (subresult)
      g_variant_unref (subresult);
  if (error)
      g_error_free (error);

  if (collection_proxy && !is_locked)
      return collection_proxy;
  else if (collection_proxy)
      g_object_unref (collection_proxy);

  return NULL;
}

/**
 * create_attributes:
 * @printer_uri: URI for the printer
 * @additional_labels: Optional labels for additional attributes
 * @additional_attrs: Optional additional attributes
 *
 * Creates a GVariant dictionary with key / value pairs that
 * can be used to identify a secret item.
 *
 * Returns: A GVariant dictionary of string pairs or NULL on error.
 */
static GVariant *
create_attributes (const gchar *printer_uri,
                   gchar **additional_attrs,
                   gchar **additional_labels)
{
  if (!printer_uri)
    {
      GTK_NOTE (PRINTING,
                g_print ("create_attributes called with invalid parameters.\n"));
      return NULL;
    }

  GVariantBuilder *const attr_builder = g_variant_builder_new (G_VARIANT_TYPE_DICTIONARY);
  /* The printer uri is the main identifying part */
  g_variant_builder_add (attr_builder, "{ss}", "uri", printer_uri);

  if (additional_labels)
    {
      int i = 0;
      for (i = 0; i < g_strv_length (additional_labels); i++)
        {
          g_variant_builder_add (attr_builder, "{ss}",
                                 additional_labels[i],
                                 additional_attrs[i]);
        }
    }

  return g_variant_builder_end(attr_builder);
}
/**
 * query_secrets_service:
 * @auth_info_required: Auth info needed for the job
 * @printer_uri: The URI of the printer used in the lookup
 *
 * Checks if a secrets service as described by the secrets-service standard
 * is available and if so it tries to find the authentication info in the
 * default collection of the service.
 *
 * Returns: Array of auth_info strings in same order as in auth_info_required.
 * If no secrets service is available or an error occured this function
 * returns NULL.
 *
 * See: http://standards.freedesktop.org/secret-service/ for documentation
 * of the used API.
 */
gchar **
query_secrets_service (gchar **auth_info_required, const gchar *printer_uri)
{
  gchar **auth_info = NULL;
  GError *error = NULL;
  GDBusConnection *session_connection = NULL;
  GDBusProxy *service_proxy = NULL,
             *session_proxy = NULL;
  GVariant *result = NULL,
           *subresult = NULL,
           *secrets_attrs = NULL;
  int i = 0;
  gsize j = 0;
  gsize length = g_strv_length (auth_info_required);
  const gboolean is_locked;

  if (!auth_info_required || !printer_uri)
    {
      GTK_NOTE (PRINTING,
                g_print ("Invalid parameters for query_secrets_service.\n"));
      return NULL;
    }

  if (!open_session (&session_connection, &service_proxy, &session_proxy))
    {
      GTK_NOTE (PRINTING, g_print ("Failed to open session.\n"));
      return NULL;
    }
  secrets_attrs = create_attributes (printer_uri, NULL, NULL);

  if (!secrets_attrs)
    {
      GTK_NOTE (PRINTING, g_print ("Failed to create attributes.\n"));
      goto done;
    }

  result = g_dbus_proxy_call_sync (service_proxy,
                                   "SearchItems",
                                   g_variant_new("(@a{ss})", secrets_attrs),
                                   G_DBUS_CALL_FLAGS_NONE,
                                   SECRETS_TIMEOUT,
                                   NULL,
                                   &error);
  if (!result)
    {
      GTK_NOTE (PRINTING, g_print ("Item Search failed: %s.\n",
                                   error->message));
      goto done;
    }

  /* Result should be two Array<ObjectPath> variants */
  const gsize array_cnt = g_variant_n_children(result);

  /* Check for an unlocked item get it and be done */
  gchar const **items = NULL;
  for (j = 0; j < array_cnt; j++)
    {
      GVariant *const item_paths = g_variant_get_child_value (result, j);
      if (!item_paths)
        {
          GTK_NOTE (PRINTING,
                    g_print ("SearchItems returned invalid result.\n"));
          continue;
        }

      items = g_variant_get_objv (item_paths, NULL);

      if (!*items)
        {
          g_variant_unref (item_paths);
          continue;
        }

      /* Access the first found item. */
      GDBusProxy *item_proxy = g_dbus_proxy_new_sync (session_connection,
                                                      G_DBUS_PROXY_FLAGS_NONE,
                                                      NULL,
                                                      SECRETS_BUS,
                                                      *items,
                                                      SECRETS_IFACE("Item"),
                                                      NULL,
                                                      &error);
      g_free ((gpointer) items);
      g_variant_unref (item_paths);

      if (!item_proxy)
        {
          GTK_NOTE (PRINTING,
                    g_print ("Failed to obtain item interface: %s\n",
                             error->message));
          continue;
        }

      GVariant *const locked = g_dbus_proxy_get_cached_property (item_proxy,
                                                                 "Locked");
      if (!locked)
        {
          GTK_NOTE (PRINTING, g_print ("Failed to lookup locked property\n"));
          continue;
        }
      g_variant_get (locked, "b", &is_locked);
      g_variant_unref (locked);

      if (is_locked)
        {
          /* Try to unlock the default collection */
          GDBusProxy *default_collection =
              get_default_collection_unlocked (session_connection,
                                               service_proxy);
          if (default_collection)
            {
              g_object_unref (default_collection);
            }
          else
              GTK_NOTE (PRINTING, g_print ("Failed to unlock collection.\n"));
        }

      GVariant *const attributes = g_dbus_proxy_get_cached_property (item_proxy,
                                                                     "Attributes");
      if (!attributes)
        {
          GTK_NOTE (PRINTING, g_print ("Failed to lookup attributes.\n"));
          continue;
        }

      GVariantIter * iter = NULL;
      g_variant_get (attributes, "a{ss}", &iter);

      /* Construct the return value */
      auth_info = g_new0 (gchar *, length);
      gchar *key = NULL;
      gchar *value = NULL;
      while (g_variant_iter_loop (iter, "{ss}", &key, &value))
        {
          /* Match attributes with required auth info */
          for (i = 0; i < length; i++)
            {
              if ((strcmp (key, "user") == 0 ||
                   strcmp (key, "username") == 0) &&
                  strcmp (auth_info_required[i], "username") == 0)
                  auth_info[i] = g_strdup (value);
              else if (strcmp (key, "domain") == 0 &&
                       strcmp (auth_info_required[i], "domain") == 0)
                  auth_info[i] = g_strdup (value);
              else if ((strcmp (key, "hostname") == 0 ||
                        strcmp (key, "server") == 0 ) &&
                       strcmp (auth_info_required[i], "hostname") == 0)
                  auth_info[i] = g_strdup (value);
            }
        }
      for (i = 0; i < length; i++)
        {
          if (strcmp (auth_info_required[i], "password") == 0)
            {
              gsize len = 0;
              /* Finally obtain the password */
              GVariant *const s_result =
                  g_dbus_proxy_call_sync (item_proxy,
                                          "GetSecret",
                                          g_variant_new("(o)",
                                              g_dbus_proxy_get_object_path(session_proxy)),
                                          G_DBUS_CALL_FLAGS_NONE,
                                          SECRETS_TIMEOUT,
                                          NULL,
                                          &error);
              if (!s_result)
                {
                  GTK_NOTE (PRINTING,
                            g_print ("Failed to get secret: %s\n", error->message));
                  g_error_free (error);
                  error = NULL;
                  continue;
                }

              GVariant *const secret = g_variant_get_child_value(s_result, 0);
              if (!secret || g_variant_n_children(secret) != 4)
                {
                  GTK_NOTE (PRINTING, g_print ("Get secret response invalid.\n"));
                  g_variant_unref (s_result);
                  g_variant_unref (secret);
                  continue;
                }

              GVariant *const s_value = g_variant_get_child_value (secret, 2);
              gconstpointer ba_passwd = g_variant_get_fixed_array(s_value,
                      &len, sizeof(guchar));

              if (!ba_passwd || strlen(ba_passwd) > len + 1)
                  GTK_NOTE (PRINTING, g_print ("Invalid secret.\n"));
              else
                  auth_info[i] = g_strdup (ba_passwd);

              g_variant_unref (s_value);
              g_variant_unref (s_result);
              g_variant_unref (secret);
              break;
            }
        }
      g_variant_iter_free (iter);
      iter = NULL;
      g_variant_unref (attributes);
    }

  if (auth_info)
    {
      gboolean success = 1;
      /* Check if we have everything */
      for (i = 0; i < length; i++)
        {
          if (!auth_info[i])
            {
              GTK_NOTE (PRINTING,
                        g_print ("Failed to find auth_info_part: %s\n",
                                 auth_info_required[i]));
              success = 0;
            }
        }
      if (!success)
        {
          g_strfreev (auth_info);
          auth_info = NULL;
        }
    }

done:
  /* Cleanup and exit */
  if (result)
      g_variant_unref (result);
  if (subresult)
      g_variant_unref (subresult);
  if (error)
      g_error_free (error);
  close_session (session_connection, service_proxy, session_proxy);
  return auth_info;
}

/**
 * secrets_service_store:
 * @auth_info: Auth info that should be stored
 * @printer_uri: URI of the printer
 *
 * Tries to store the auth_info in a secrets service.
 *
 * Returns: FALSE on failure.
 */
gboolean
secrets_service_store (gchar **auth_info,
                       gchar **auth_info_labels,
                       const gchar* printer_uri)
{
  GError *error = NULL;
  GDBusConnection *session_connection = NULL;
  GDBusProxy *service_proxy = NULL,
             *session_proxy = NULL,
             *default_collection = NULL;
  GVariantBuilder *prop_builder = NULL;
  GVariant *attributes = NULL,
           *properties = NULL,
           *secret = NULL,
           *result = NULL;
  gchar *password = NULL;
  gboolean success = FALSE;
  gchar **additional_attrs = NULL;
  gchar **additional_labels = NULL;
  int i = 0, additional_count = 0, length = 0;

  if (!open_session (&session_connection, &service_proxy, &session_proxy))
    {
      GTK_NOTE (PRINTING, g_print ("Failed to open dbus session.\n"));
      return FALSE;
    }
  default_collection = get_default_collection_unlocked (session_connection,
                                                        service_proxy);

  if (!default_collection)
    {
      GTK_NOTE (PRINTING, g_print ("Failed to get default collection.\n"));
      goto done;
    }

  length = g_strv_length (auth_info_labels);

  additional_attrs = g_new0 (gchar*, length + 1);
  additional_labels = g_new0 (gchar*, length + 1);
  /* The labels user and server are chosen to be compatible with
     the attributes used by system-config-printer */
  for (i = 0; i < g_strv_length (auth_info_labels); i++)
    {
      if (g_strcmp0 (auth_info_labels[i], "username") == 0)
        {
          additional_attrs[additional_count] = auth_info[i];
          additional_labels[additional_count++] = "user";
        }
      else if (g_strcmp0 (auth_info_labels[i], "hostname") == 0)
        {
          additional_attrs[additional_count] = auth_info[i];
          additional_labels[additional_count++] = "server";
        }
      else if (g_strcmp0 (auth_info_labels[i], "password") == 0)
        {
          password = auth_info[i];
        }
    }

  attributes = create_attributes (printer_uri, additional_attrs, additional_labels);
  g_free (additional_labels);
  g_free (additional_attrs);
  if (!attributes)
    {
      GTK_NOTE (PRINTING, g_print ("Failed to create attributes.\n"));
      goto done;
    }

  if (!password)
    {
      GTK_NOTE (PRINTING, g_print ("No secret to store.\n"));
      goto done;
    }

  prop_builder = g_variant_builder_new (G_VARIANT_TYPE_DICTIONARY);

  g_variant_builder_add (prop_builder, "{sv}", SECRETS_IFACE("Item.Label"),
                         g_variant_new_string(printer_uri));
  g_variant_builder_add (prop_builder, "{sv}", SECRETS_IFACE("Item.Attributes"),
                         attributes);

  properties = g_variant_builder_end (prop_builder);

  secret = g_variant_new ("(oay@ays)",
                          g_dbus_proxy_get_object_path (session_proxy),
                          NULL,
                          g_variant_new_bytestring (password),
                          "text/plain");

  result = g_dbus_proxy_call_sync (default_collection,
                                   "CreateItem",
                                   g_variant_new ("(@a{sv}@(oayays)b)",
                                                  properties,
                                                  secret,
                                                  TRUE),
                                   G_DBUS_CALL_FLAGS_NONE,
                                   SECRETS_TIMEOUT,
                                   NULL,
                                   &error);

  if (!result)
    {
      GTK_NOTE (PRINTING,
                g_print ("Failed to create collection: %s\n", error->message));
      g_error_free (error);
    }
  else
    {
       gchar *item = NULL;
       g_variant_get (result, "(&o&o)", &item, NULL);
       if (item && strlen (item) > 1)
           success = TRUE;
       g_variant_unref (result);
    }

done:
  if (default_collection)
      g_object_unref (default_collection);

  close_session (session_connection, service_proxy, session_proxy);
  return success;
}
